"""If seq is the same in a short time, the device will not respond.
Increment seq by 100 won't help, because seq may still be the same.
"""

import random

import miio
from miio.miioprotocol import MiIOProtocol


class MiIOProtocolFix(MiIOProtocol):
    discovered = False
    seq = -1

    @property
    def _discovered(self):
        return self.discovered

    @_discovered.setter
    def _discovered(self, value: bool):
        self.discovered = value

        if not value:
            self.seq = random.randint(0, 999999)

    @property
    def _id(self) -> int:
        self.seq += 1
        if self.seq >= 999999:
            self.seq = 1
        return self.seq


class Device(miio.Device):
    # noinspection PyMissingConstructor
    def __init__(self, ip: str, token: str):
        self._protocol = MiIOProtocolFix(ip, token)

    def get_device_list(self) -> list:
        devices = {}

        for _ in range(16):
            # load only 8 device per part
            part = self.send("get_device_list")
            if len(part) == 0:
                return []

            for item in part:
                devices[item["num"]] = item

        return list(devices.values())

    def get_device_prop(self, did: str, params: dict) -> dict:
        if did.startswith("lumi."):
            values = self.send("get_device_prop", [did] + list(params.keys()))
            names = params.values()
            return dict(zip(names, values))

        else:
            # [{'code': 0, 'did': '123', 'piid': 1, 'siid': 3, 'value': 123}]
            values = self.send(
                "get_properties",
                [
                    {"did": did, "siid": int(p[0]), "piid": int(p[2])}
                    for p in params.keys()
                ],
            )
            return {params[f"{p['siid']}.{p['piid']}"]: p["value"] for p in values}



from ..generic import FhemModule
import asyncio
import time

from .. import fhem
from ..xiaomi_gateway3 import xiaomi_gateway3
from .. import fhem_pythonbinding as fhepy

attr_settings = {
    "lumi.sensor_magnet.v2": {
        "devStateIcon": "1:fts_door_open\@red 0:fts_door\@green",
        "stateFormat": "contact",
        "icon": "tuer_fenster_kontakt",
    },
    "lumi.sensor_ht.v1": {
        "stateFormat": "temperature °C, humidity %",
        "icon": "temp_temperature",
    },
    "lumi.sensor_ht.v2": {
        "stateFormat": "temperature °C, humidity %, pressure kPa",
        "icon": "temp_temperature",
    },
    "lumi.sensor_motion.v1": {
        "devStateIcon": "motion:motion_detector\@red off:motion_detector\@green no_motion:motion_detector\@green",
        "icon": "people_sensor",
    },
    "lumi.gateway.mgl03": {
        "stateFormat": "presence",
        "devStateIcon": "online:it_wifi\@green offline:it_wifi\@red",
        "icon": "tradfri_gateway",
    },
}


class xiaomi_gateway3_device(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.gateway = None
        self.device_details = None
        self.last_update = 0
        self.loop = asyncio.get_event_loop()
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        self.hash = hash

        if len(args) < 5:
            return "Usage: define devname PythonModule xiaomi_gateway3_device <GATEWAY_NAME> <DID>"

        self.gw_name = args[3]
        self.did = args[4]

        hash["GATEWAY"] = self.gw_name
        hash["DID"] = self.did

        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)

        self.create_async_task(self.connect_gw())
        self.offline_check_task = self.create_async_task(self.offline_check())

        return ""

    async def offline_check(self):
        while True:
            if time.time() - self.last_update > 3700:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "offline", 1
                )
            await asyncio.sleep(300)

    async def connect_gw(self):
        while self.gateway is None:
            self.gateway = fhepy.getFhemPyDeviceByName(self.gw_name)
            if self.gateway:
                try:
                    await self.gateway.register_device(self, self.update)
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "online", 1
                    )
                except:
                    self.gateway = None
                    pass
            else:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", f"gateway {self.gw_name} not found", 1
                )
            await asyncio.sleep(10)

    async def initialize(self, device):
        # first update, set attributes and device readings like model, sid, ...
        if self.device_details is None:
            self.device_details = device
            if self.device_details["model"] in attr_settings:
                for attr in attr_settings[self.device_details["model"]]:
                    if await fhem.AttrVal(self.hash["NAME"], attr, "") == "":
                        await fhem.CommandAttr(
                            self.hash,
                            f"{self.hash['NAME']} {attr} {attr_settings[self.device_details['model']][attr]}",
                        )
            for reading in self.device_details:
                if reading == "init":
                    await self.update(self.device_details["init"])
                    continue
                elif reading == "params":
                    continue
                else:
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, reading, self.device_details[reading], 1
                    )

    async def update(self, data):
        self.logger.debug(f"update call {str(data)}")
        self.last_update = time.time()
        if data is None:
            return

        await fhem.readingsBeginUpdate(self.hash)
        # device is online
        await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "online")
        # update data
        for reading in data:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, reading.replace(" ", "_"), str(data[reading])
            )
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio
import aiohttp
import sys

from async_upnp_client.aiohttp import AiohttpSessionRequester
from async_upnp_client.advertisement import UpnpAdvertisementListener
from async_upnp_client import UpnpFactory
from async_upnp_client.search import async_search as async_ssdp_search
from ipaddress import IPv4Address


# ssdp search class can be used by other modules as well
class ssdp:

    __instance = None

    @staticmethod
    def getInstance(logger):
        if ssdp.__instance is None:
            ssdp.__instance = ssdp(logger)
        return ssdp.__instance

    def __init__(self, logger):
        if ssdp.__instance is not None:
            raise Exception("ssdp already defined, use getInstance")
        self.logger = logger
        self.listeners = []
        self.listener = None
        self.search_task = None
        self.advertisement_task = None
        self.nr_started_searches = 0

        # build upnp/aiohttp requester
        self.session = aiohttp.ClientSession()
        self.requester = AiohttpSessionRequester(self.session, True)
        # create upnp device
        self.factory = UpnpFactory(self.requester)

    async def start_search(self):
        self.nr_started_searches += 1
        self.search_task = asyncio.create_task(self.search())

        if self.advertisement_task is None:
            self.advertisement_task = asyncio.create_task(self.advertisements())

    async def stop_search(self):
        self.nr_started_searches -= 1
        # stop search only when last client stops it
        if self.nr_started_searches == 0:
            await self.session.close()
            if self.search_task is not None:
                self.search_task.cancel()
            if self.advertisement_task is not None:
                self.advertisement_task.cancel()
            if self.listener is not None:
                await self.listener.async_stop()

    def register_listener(self, listener, ssdp_filter={"service_type": "ssdp:all"}):
        listenerFilter = {
            "listener": listener,
            "ssdp_filter": ssdp_filter,
            "found_devices": {},
        }
        self.listeners.append(listenerFilter)

    async def updated_device(self, udn):
        return

    async def create_device(self, msg, data):
        upnp_device = None
        try:
            if msg != "byebye":
                upnp_device = await self.factory.async_create_device(data["location"])
        except:
            # upnp_device remains None
            pass
        return upnp_device

    async def handle_msg(self, msg, data):
        try:
            data = {key.lower(): str(value) for key, value in data.items()}
            usn = data["usn"]
            arr = usn.split("::")
            udn = arr[0]
            if len(arr) > 1:
                st = arr[1]
            else:
                st = ""
            self.logger.debug("found: " + usn)
            # filter for listeners
            for listenerFilter in self.listeners:
                listener = listenerFilter["listener"]
                ssdp_filter = listenerFilter["ssdp_filter"]
                filter_udn = None
                filter_st = None
                if "udn" in ssdp_filter:
                    filter_udn = ssdp_filter["udn"]
                if "service_type" in ssdp_filter:
                    filter_st = ssdp_filter["service_type"]
                if (udn == filter_udn or filter_udn is None) and (
                    st == filter_st or filter_st == "ssdp:all"
                ):
                    if usn not in listenerFilter["found_devices"] and msg == "alive":
                        self.logger.debug("create device: " + usn)
                        listenerFilter["found_devices"][usn] = await self.create_device(
                            msg, data
                        )
                        if listenerFilter["found_devices"][usn]:
                            self.logger.debug("found device: " + usn)
                            await listener.found_device(
                                listenerFilter["found_devices"][usn]
                            )
                    elif usn in listenerFilter["found_devices"] and msg == "byebye":
                        self.logger.debug("removed device: " + usn)
                        await listener.removed_device(
                            listenerFilter["found_devices"][usn]
                        )
                        del listenerFilter["found_devices"][usn]
        except:
            self.logger.exception("Error in handle_msg")

    async def search(self):
        try:
            await self.search_once()
        except asyncio.CancelledError:
            pass

    async def search_once(self):
        timeout = 30
        service_type = "ssdp:all"
        source_ip = None
        if sys.platform == "win32" and not source_ip:
            self.logger.debug(
                'Running on win32 without --bind argument, forcing to "0.0.0.0"'
            )
            source_ip = (
                "0.0.0.0"  # force to IPv4 to prevent asyncio crash/WinError 10022
            )
        if source_ip:
            source_ip = IPv4Address(source_ip)

        async def on_response(data):
            await self.handle_msg("alive", data)

        try:
            await async_ssdp_search(
                service_type=service_type,
                source_ip=source_ip,
                timeout=timeout,
                async_callback=on_response,
            )
        except asyncio.CancelledError:
            self.logger.debug("search cancelled")

    async def advertisements(self):
        """Listen for advertisements."""
        try:
            source_ip = None
            if sys.platform == "win32" and not source_ip:
                self.logger.debug(
                    'Running on win32 without --bind argument, forcing to "0.0.0.0"'
                )
                source_ip = (
                    "0.0.0.0"  # force to IPv4 to prevent asyncio crash/WinError 10022
                )
            if source_ip:
                source_ip = IPv4Address(source_ip)

            async def on_alive(data):
                await self.handle_msg("alive", data)

            async def on_byebye(data):
                await self.handle_msg("byebye", data)

            async def on_update(data):
                return

            self.listener = UpnpAdvertisementListener(
                on_alive=on_alive,
                on_byebye=on_byebye,
                on_update=on_update,
                source_ip=source_ip,
            )
            await self.listener.async_start()
        except:
            self.logger.error("Error in advertisements listener")



import asyncio
import functools
import concurrent.futures
import time
import random
import logging

from enum import IntEnum
from datetime import datetime

from bluepy.btle import BTLEException
from . import eq3btsmart as eq3
from .connection import BTLEConnection
from dbus import DBusException

from .. import utils
from .. import fhem
from ..generic import FhemModule


class Mode(IntEnum):
    """ Thermostat modes. """

    Unknown = -1
    Closed = 0
    Open = 1
    Auto = 2
    Manual = 3
    Away = 4
    Boost = 5


# TODO set schedules
# TODO set windowOpen, windowOpenTime, eco/comfortTemperature


class eq3bt(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        attr_list = {
            "keep_connected": {
                "default": "on",
                "format": "str",
                "options": "on,off",
                "help": "On...keeps bluetooth low energy connection active which makes commands to be executed immediately",
            }
        }
        self.set_attr_config(attr_list)

        set_list_conf = {
            "on": {},
            "off": {},
            "desiredTemperature": {
                "args": ["target_temp"],
                "options": "slider,4.5,0.5,30,1",
            },
            "updateStatus": {},
            "boost": {
                "args": ["target_state"],
                "options": "on,off",
                "help": "Activate boost for 300s",
            },
            "mode": {"args": ["target_mode"], "options": "manual,automatic"},
            "eco": {},
            "comfort": {},
            "childlock": {"args": ["target_state"], "options": "on,off"},
            "resetConsumption": {
                "args": ["cons_var"],
                "options": "all,consumption,consumptionToday,consumptionYesterday",
            },
        }
        self.set_set_config(set_list_conf)

        self._last_update = 0
        self._mac = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        self.hash = hash
        if len(args) < 4:
            return "Usage: define eq3_livingroom PythonModule eq3bt <MAC>"
        self._mac = args[3]
        self.hash["MAC"] = self._mac
        self.logger.info(f"Define: eq3bt {self._mac}")

        icon = await fhem.AttrVal(self.hash["NAME"], "icon", "noicon")
        if icon == "noicon":
            await fhem.CommandAttr(
                self.hash, self.hash["NAME"] + " icon sani_heating_temp"
            )
        await fhem.readingsSingleUpdateIfChanged(self.hash, "presence", "offline", 1)
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connecting", 1)

        # handle missing dbus configuration
        try:
            self.thermostat = FhemThermostat(
                self.logger,
                self._mac,
                keep_connection=self._attr_keep_connected == "on",
            )
        except DBusException:
            dbus_conf_err = (
                "Please add following configuration to /etc/dbus-1/system.d/bluetooth.conf:\n"
                '<policy user="fhem">\n'
                '    <allow own="org.bluez"/>\n'
                '    <allow send_destination="org.bluez"/>\n'
                '    <allow send_interface="org.bluez.GattCharacteristic1"/>\n'
                '    <allow send_interface="org.bluez.GattDescriptor1"/>\n'
                '    <allow send_interface="org.freedesktop.DBus.ObjectManager"/>\n'
                '    <allow send_interface="org.freedesktop.DBus.Properties"/>\n'
                "</policy>\n\n"
                "ATTENTION: On remote device change the user account above to the one which runs fhempy (e.g. pi)\n\n"
                "Restart dbus afterwards: sudo systemctl restart dbus"
            )
            self.logger.error(dbus_conf_err)
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "state", dbus_conf_err, 1
            )
            return dbus_conf_err

        self.create_async_task(self.check_online())
        self.create_async_task(self.consumption_rotate())
        return ""

    def seconds_till_midnight(self):
        """Get the number of seconds until midnight."""
        n = datetime.now()
        return (
            ((24 - n.hour - 1) * 60 * 60) + ((60 - n.minute - 1) * 60) + (60 - n.second)
        )

    async def consumption_rotate(self):
        while True:
            await asyncio.sleep(self.seconds_till_midnight())
            consumption = float(
                await fhem.ReadingsVal(self.hash["NAME"], "consumptionToday", "0")
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionYesterday", consumption, 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionToday", "0", 1
            )

    async def set_resetConsumption(self, hash, params):
        cons_var = params["cons_var"]
        if cons_var == "all":
            await fhem.readingsSingleUpdateIfChanged(self.hash, "consumption", 0, 1)
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionYesterday", 0, 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "consumptionToday", 0, 1
            )
        else:
            await fhem.readingsSingleUpdateIfChanged(self.hash, cons_var, 0, 1)

    async def set_attr_keep_connected(self, hash):
        self.thermostat.set_keep_connected(self._attr_keep_connected == "on")

    async def check_online(self):
        waittime = 300
        if self._attr_keep_connected == "on":
            waittime = 60
        await asyncio.sleep(int(random.random() * 100))
        while True:
            try:
                if time.time() - self._last_update > (waittime * 4):
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "presence", "offline", 1
                    )
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "update", 1
                    )
                await self.update_all()
            except:
                self.logger.error(f"Failed to update, retry in {waittime}s")
            await asyncio.sleep(waittime)

    async def update_all(self):
        self.logger.debug("start update_all")
        with concurrent.futures.ThreadPoolExecutor() as pool:
            await asyncio.get_event_loop().run_in_executor(
                pool, functools.partial(self.thermostat.update_all)
            )
        await self.update_all_readings()

    async def update_all_readings(self):
        await self.update_readings()
        await self.update_id_readings()
        await self.update_schedule_readings()

    async def update_readings(self):
        old_valve_pos = float(
            await fhem.ReadingsVal(self.hash["NAME"], "valvePosition", "0")
        )
        old_consumption = float(
            await fhem.ReadingsVal(self.hash["NAME"], "consumption", "0")
        )
        old_consumption_today = float(
            await fhem.ReadingsVal(self.hash["NAME"], "consumptionToday", "0")
        )
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "battery", self.thermostat.battery
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "boost", self.thermostat.boost
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "childlock", self.thermostat.locked
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "desiredTemperature", self.thermostat.target_temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "ecoTemperature", self.thermostat.eco_temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "temperatureOffset", self.thermostat.temperature_offset
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "comfortTemperature", self.thermostat.comfort_temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "mode", self.thermostat.fhem_mode
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "state", self.thermostat.state
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "completeState", self.thermostat.mode_readable
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "valvePosition", self.thermostat.valve_state
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "awayEnd", self.thermostat.away_end
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "windowOpen", self.thermostat.window_open
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "windowOpenTemperature", self.thermostat.window_open_temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "windowOpenTime", self.thermostat.window_open_time
        )
        await fhem.readingsBulkUpdateIfChanged(self.hash, "presence", "online")
        if (time.time() - self._last_update) < 400:
            consumption_diff = (
                (old_valve_pos + self.thermostat.valve_state)
                / 2
                / 100
                * (time.time() - self._last_update)
                / 60
            )
        else:
            consumption_diff = 0
        new_consumption = round(old_consumption + consumption_diff, 2)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "consumption", new_consumption
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash,
            "consumptionToday",
            round(old_consumption_today + consumption_diff, 2),
        )
        await fhem.readingsEndUpdate(self.hash, 1)
        self._last_update = time.time()

    async def update_id_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "firmware", self.thermostat.firmware_version
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "serialNumber", self.thermostat.device_serial
        )
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_schedule_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        for day in self.thermostat.schedule.keys():
            reading = f"schedule_{day}_1"
            if self.thermostat.schedule[day].base_temp == 0 or isinstance(
                self.thermostat.schedule[day].next_change_at, int
            ):
                await fhem.readingsBulkUpdateIfChanged(self.hash, reading, "-")
                last_change = "00:00"
            else:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash,
                    reading,
                    f"00:00 - {self.thermostat.schedule[day].next_change_at.strftime('%H:%M')}: {self.thermostat.schedule[day].base_temp}",
                )
                last_change = self.thermostat.schedule[day].next_change_at.strftime(
                    "%H:%M"
                )
            last_schedule = False
            for h in range(0, 6):
                reading = f"schedule_{day}_{h+2}"
                if (
                    h == 6
                    or self.thermostat.schedule[day].hours[h].target_temp == 0
                    or isinstance(
                        self.thermostat.schedule[day].hours[h].next_change_at, int
                    )
                    or last_schedule
                ):
                    if last_schedule:
                        await fhem.readingsBulkUpdateIfChanged(self.hash, reading, "-")
                    else:
                        value = f"{last_change} - 00:00: {self.thermostat.schedule[day].base_temp}"
                        await fhem.readingsBulkUpdateIfChanged(
                            self.hash, reading, value
                        )
                    last_schedule = True
                else:
                    value = f"{last_change} - {self.thermostat.schedule[day].hours[h].next_change_at.strftime('%H:%M')}: {self.thermostat.schedule[day].hours[h].target_temp}"
                    last_change = (
                        self.thermostat.schedule[day]
                        .hours[h]
                        .next_change_at.strftime("%H:%M")
                    )
                    await fhem.readingsBulkUpdateIfChanged(self.hash, reading, value)
        await fhem.readingsEndUpdate(self.hash, 1)

    async def set_and_update(self, fct):
        await utils.run_blocking(fct)
        await self.update_readings()

    # SET Functions BEGIN
    async def set_on(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(self.thermostat.set_target_temperature, 30)
            )
        )

    async def set_off(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(self.thermostat.set_target_temperature, 4.5)
            )
        )

    async def set_desiredTemperature(self, hash, params):
        temp = float(params["target_temp"])
        self.create_async_task(
            self.set_and_update(
                functools.partial(self.thermostat.set_target_temperature, temp)
            )
        )

    async def set_updateStatus(self, hash, params):
        self.create_async_task(self.update_all())

    async def set_boost(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_boost, params["target_state"] == "on"
                )
            )
        )

    async def set_mode(self, hash, params):
        target_mode = params["target_mode"]
        if target_mode == "automatic":
            target_mode = eq3.Mode.Auto
        else:
            target_mode = eq3.Mode.Manual
        self.create_async_task(
            self.set_and_update(
                functools.partial(self.thermostat.set_fhem_mode, target_mode)
            )
        )

    async def set_eco(self, hash, params):
        self.create_async_task(
            self.set_and_update(functools.partial(self.thermostat.activate_eco))
        )

    async def set_comfort(self, hash, params):
        self.create_async_task(
            self.set_and_update(functools.partial(self.thermostat.activate_comfort))
        )

    async def set_childlock(self, hash, params):
        self.create_async_task(
            self.set_and_update(
                functools.partial(
                    self.thermostat.set_locked, params["target_state"] == "on"
                )
            )
        )

    # SET Functions END


class FhemThermostat(eq3.Thermostat):
    def __init__(self, logger, mac, keep_connection):
        self.logger = logger
        self._keep_conection = keep_connection
        super(FhemThermostat, self).__init__(mac, BTLEConnection, keep_connection=True)

    def set_keep_connection(self, new_state):
        self.set_keep_connected(new_state)

    def update_all(self):
        super().update()
        super().query_id()
        for day in range(0, 6):
            super().query_schedule(day)

    def set_target_temperature(self, temp):
        self.target_temperature = temp

    def set_boost(self, state):
        self.boost = state

    def set_locked(self, state):
        self.locked = state

    def set_fhem_mode(self, mode):
        self.mode = mode

    @property
    def fhem_mode(self):
        if self._mode == Mode.Boost:
            return "boost"
        elif self._mode == Mode.Away:
            return "away"
        elif self._mode == Mode.Closed:
            return "manual"
        elif self._mode == Mode.Open:
            return "manual"
        elif self._mode == Mode.Manual:
            return "manual"
        elif self._mode == Mode.Auto:
            return "automatic"

    @property
    def state(self):
        if self._mode == Mode.Boost:
            return "boost"
        elif self._mode == Mode.Away:
            return "away"
        elif self._mode == Mode.Closed:
            return "off"
        elif self._mode == Mode.Open:
            return "on"
        elif self._mode == Mode.Manual:
            return "manual"
        elif self._mode == Mode.Auto:
            return "automatic"

    @property
    def battery(self):
        if self.low_battery:
            return "low"
        return "ok"



import asyncio
import functools
import time
import re

from fhempy.lib.generic import FhemModule

from bluepy.btle import (
    BTLEDisconnectError,
    ScanEntry,
    Scanner,
    Peripheral,
    ADDR_TYPE_PUBLIC,
    ADDR_TYPE_RANDOM,
    BluepyHelper,
    AssignedNumbers,
)

from .. import fhem, utils


class PeripheralTimeout(Peripheral):
    def __init__(
        self, deviceAddr=None, addrType=ADDR_TYPE_PUBLIC, iface=None, timeout=None
    ):
        BluepyHelper.__init__(self)
        self._serviceMap = None  # Indexed by UUID
        (self.deviceAddr, self.addrType, self.iface) = (None, None, None)

        if isinstance(deviceAddr, ScanEntry):
            self._connect(
                deviceAddr.addr, deviceAddr.addrType, deviceAddr.iface, timeout
            )
        elif deviceAddr is not None:
            self._connect(deviceAddr, addrType, iface, timeout)

    def _connect(self, addr, addrType=ADDR_TYPE_PUBLIC, iface=None, timeout=None):
        if len(addr.split(":")) != 6:
            raise ValueError("Expected MAC address, got %s" % repr(addr))
        if addrType not in (ADDR_TYPE_PUBLIC, ADDR_TYPE_RANDOM):
            raise ValueError(
                "Expected address type public or random, got {}".format(addrType)
            )
        self._startHelper(iface)
        self.addr = addr
        self.addrType = addrType
        self.iface = iface
        if iface is not None:
            self._writeCmd("conn %s %s %s\n" % (addr, addrType, "hci" + str(iface)))
        else:
            self._writeCmd("conn %s %s\n" % (addr, addrType))
        rsp = self._getResp("stat", timeout)
        if rsp is None:
            raise BTLEDisconnectError(
                "Timed out while trying to connect to peripheral %s, addr type: %s"
                % (addr, addrType),
                rsp,
            )
        while rsp["state"][0] == "tryconn":
            rsp = self._getResp("stat", timeout)
        if rsp["state"][0] != "conn":
            self._stopHelper()
            raise BTLEDisconnectError(
                "Failed to connect to peripheral %s, addr type: %s" % (addr, addrType),
                rsp,
            )

    def connect(self, addr, addrType=ADDR_TYPE_PUBLIC, iface=None, timeout=None):
        if isinstance(addr, ScanEntry):
            self._connect(addr.addr, addr.addrType, addr.iface, timeout)
        elif addr is not None:
            self._connect(addr, addrType, iface, timeout)


class scanner:

    scanner_instance = None

    def __init__(self, logger):
        self.logger = logger
        self.loop = asyncio.get_event_loop()
        self._mac_listener = {}
        self._mac_lastfound = {}
        self._scan_task = None
        self._scan_interval = 10
        self._scan_duration = 10
        self._iface = 0

    def get_instance(logger):
        if scanner.scanner_instance is None:
            scanner.scanner_instance = scanner(logger)
        return scanner.scanner_instance

    def register_mac_listener(self, mac, listener):
        self._mac_listener[mac.lower()] = listener
        if self._scan_task is None:
            self._scan_task = asyncio.create_task(self.loop_scan())

    def unregister_mac_listener(self, mac):
        if mac in self._mac_listener:
            del self._mac_listener[mac]
        if len(self._mac_listener) == 0:
            if self._scan_task:
                self._scan_task.cancel()
                self._scan_task = None

    def handleDiscovery(self, dev, isNewDev, isNewData):
        asyncio.run_coroutine_threadsafe(
            self.handle_discovery(dev, isNewDev, isNewData), self.loop
        ).result()

    async def handle_discovery(self, dev, isNewDev, isNewData):
        if dev.addr.lower() in self._mac_listener:
            self.logger.debug(f"found {dev.addr.lower()}")
            if isNewDev:
                self._mac_lastfound[dev.addr.lower()] = time.time()
                name = ""
                if dev.getValueText(ScanEntry.SHORT_LOCAL_NAME):
                    name = dev.getValueText(ScanEntry.SHORT_LOCAL_NAME)
                else:
                    name = dev.getValueText(ScanEntry.COMPLETE_LOCAL_NAME)
                await getattr(self._mac_listener[dev.addr.lower()], "update_readings")(
                    "present", dev.addr.lower(), name, dev.rssi
                )

    def set_scan_interval(self, scan_interval):
        self._scan_interval = scan_interval

    def set_scan_duration(self, scan_duration):
        self._scan_duration = scan_duration

    def set_hci_device(self, hci_nr):
        self._iface = hci_nr

    async def loop_scan(self):
        while True:
            for mac in self._mac_lastfound:
                self._mac_lastfound[mac] = 0
            try:
                await utils.run_blocking(functools.partial(self._do_scan))
            except:
                self.logger.error("Failed to do scan")
            for mac in self._mac_listener:
                try:
                    if mac not in self._mac_lastfound or self._mac_lastfound[mac] == 0:
                        await getattr(self._mac_listener[mac], "update_readings")(
                            "absent", mac, "", 0
                        )
                    else:
                        await getattr(
                            self._mac_listener[mac], "check_update_characteristics"
                        )()
                except:
                    self.logger.exception("Failed to handle updates after scan")
            await asyncio.sleep(self._scan_interval)

    def _do_scan(self):
        self.scanner = Scanner(self._iface).withDelegate(self)
        try:
            self.scanner.scan(self._scan_duration)
        except BTLEDisconnectError:
            return
        except Exception as ex:
            self.logger.error(f"Failed to scan: {ex}")


class ble_presence(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.hash = None

        self._name = ""
        self._rssi = 0
        self._address = ""
        self._presence = ""
        self._found = False
        self._peripheral = PeripheralTimeout()
        self._last_char_update = 0
        self._hci_nr = 0

        self._attr_scan_interval = 10
        self._attr_hci_device = "hci0"
        self._attr_scan_duration = 10
        self._attr_read_characteristics = "battery"

        self._attr_list = {
            "scan_interval": {"default": 10, "format": "int"},
            "hci_device": {"default": "hci0"},
            "scan_duration": {"default": 10, "format": "int"},
            "read_characteristics": {
                "default": "battery",
                "options": "all,off,battery",
            },
        }
        self.set_attr_config(self._attr_list)

    async def update_readings(self, presence, address, name, rssi):
        if self._rssi != rssi:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "rssi", str(rssi), 1)
            self._rssi = rssi

        if (
            name is not None
            and self._name != name
            and presence != "absent"
            and name != ""
        ):
            await fhem.readingsSingleUpdateIfChanged(self.hash, "name", name, 1)
            self._name = name

        if self._presence != presence:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "presence", presence, 1)
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", presence, 1)
            self._presence = presence

    async def check_update_characteristics(self):
        if self._attr_read_characteristics == "off":
            return

        if time.time() - self._last_char_update > 7200 and self._presence == "present":
            try:
                await self.update_characteristics()
            except:
                self.logger.error("Failed to update characteristics")

    async def update_characteristics(self):
        try:
            await utils.run_blocking(
                functools.partial(
                    self._peripheral.connect,
                    self._address,
                    ADDR_TYPE_PUBLIC,
                    self._hci_nr,
                    5,
                )
            )
        except:
            return
        services = await utils.run_blocking(
            functools.partial(self._peripheral.getServices)
        )
        await fhem.readingsBeginUpdate(self.hash)
        try:
            if self._attr_read_characteristics == "all":
                for service in services:
                    for char in service.getCharacteristics():
                        if char.supportsRead():
                            reading = (
                                service.uuid.getCommonName().replace(" ", "")
                                + "_"
                                + char.uuid.getCommonName().replace(" ", "")
                            )
                            if char.uuid.getCommonName() == "Battery Level":
                                reading = "battery"
                            try:
                                val = await utils.run_blocking(
                                    functools.partial(char.read)
                                )
                            except:
                                val = "failed"
                            try:
                                if len(val) == 1:
                                    val = ord(val)
                                else:
                                    val = val.decode("ascii")
                            except:
                                val = val.hex()
                                pass
                            await fhem.readingsBulkUpdateIfChanged(
                                self.hash, reading, val
                            )
            else:
                service = self._peripheral.getServiceByUUID(
                    AssignedNumbers.batteryService
                )
                char = service.getCharacteristics(AssignedNumbers.batteryLevel)
                batt = ord(char[0].read())
                await fhem.readingsBulkUpdateIfChanged(self.hash, "battery", batt)
            self._last_char_update = time.time()
        except:
            self.logger.error("Failed to get charachteristics")
        self._peripheral.disconnect()
        await fhem.readingsEndUpdate(self.hash, 1)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 4:
            return "Usage: define p_mysmartphone PythonModule ble_presence <MAC>"

        self._address = args[3]
        self.hash["MAC"] = args[3]

        scanner.get_instance(self.logger).register_mac_listener(self._address, self)

    async def set_attr_scan_interval(self, hash):
        scanner.get_instance(self.logger).set_scan_interval(self._attr_scan_interval)

    async def set_attr_scan_duration(self, hash):
        scanner.get_instance(self.logger).set_scan_duration(self._attr_scan_duration)

    async def set_attr_hci_device(self, hash):
        self._hci_nr = re.search(r"\d+$", self._attr_hci_device)[0]
        scanner.get_instance(self.logger).set_hci_device(self._hci_nr)

    # FHEM FUNCTION
    async def Undefine(self, hash):
        scanner.get_instance(self.logger).unregister_mac_listener(self._address)
        await super().Undefine(hash)



import asyncio
import logging
import pytest
import os
from fhempy.lib.pkg_installer import check_and_install_dependencies
from tests.utils import mock_fhem


@pytest.mark.asyncio
async def test_everything(mocker):
    # prepare
    mock_fhem.mock_module(mocker)
    testhash = {"NAME": "testdevice"}
    await check_and_install_dependencies("googlecast")
    from fhempy.lib.googlecast.googlecast import googlecast

    fhempy_device = googlecast(logging.getLogger(__name__))
    await fhempy_device.Define(
        testhash,
        ["testdevice", "PythonModule", "googlecast", os.environ["CAST_DEVICE_NAME"]],
        {},
    )
    assert mock_fhem.readings["testdevice"]["state"] == "offline"

    while fhempy_device.cast is None:
        await asyncio.sleep(1)

    # quitApp
    await fhempy_device.Set(
        testhash,
        ["testdevice", "quitApp"],
        {},
    )
    await asyncio.sleep(10)
    assert mock_fhem.readings["testdevice"]["display_name"] == "Backdrop"

    # check if readings are there
    assert "state" in mock_fhem.readings["testdevice"]

    # set spotify attributes
    await fhempy_device.Attr(
        testhash, ["set", "testdevice", "spotify_sp_dc", os.environ["SP_DC"]], {}
    )
    await fhempy_device.Attr(
        testhash, ["set", "testdevice", "spotify_sp_key", os.environ["SP_KEY"]], {}
    )
    await asyncio.sleep(10)
    assert (
        mock_fhem.readings["testdevice"]["spotify_user"]
        != "attr spotify_sp... required"
    )

    # play spotify
    await fhempy_device.Set(
        testhash,
        [
            "testdevice",
            "play",
            "https://open.spotify.com/artist/4PBCFEjR4a3OGdOZ6jeKKM",
        ],
        {},
    )
    await asyncio.sleep(40)
    assert mock_fhem.readings["testdevice"]["display_name"] == "Spotify"

    # quitApp
    await fhempy_device.Set(
        testhash,
        ["testdevice", "quitApp"],
        {},
    )
    await asyncio.sleep(10)
    assert mock_fhem.readings["testdevice"]["display_name"] == "Backdrop"

    # display website
    await fhempy_device.Set(testhash, ["testdevice", "displayWebsite"], {})
    await asyncio.sleep(15)
    assert mock_fhem.readings["testdevice"]["display_name"] == "DashCast"
    assert mock_fhem.readings["testdevice"]["status_text"] == "Home of FHEM"

    # play youtube
    await fhempy_device.Set(
        testhash,
        ["testdevice", "play", "https://www.youtube.com/watch?v=fCZVL_8D048"],
        {},
    )
    await asyncio.sleep(30)
    assert mock_fhem.readings["testdevice"]["display_name"] == "YouTube"
    assert mock_fhem.readings["testdevice"]["mediaPlayerState"] == "PLAYING"

    # check if position is updated
    await asyncio.sleep(45)
    assert mock_fhem.readings["testdevice"]["mediaCurrentPosition"] >= 30

    # pause
    await fhempy_device.Set(
        testhash,
        ["testdevice", "pause"],
        {},
    )
    await asyncio.sleep(5)
    assert mock_fhem.readings["testdevice"]["mediaPlayerState"] == "PAUSED"

    # play
    await fhempy_device.Set(
        testhash,
        ["testdevice", "play"],
        {},
    )
    await asyncio.sleep(5)
    assert mock_fhem.readings["testdevice"]["mediaPlayerState"] == "PLAYING"

    # quitApp
    await fhempy_device.Set(
        testhash,
        ["testdevice", "quitApp"],
        {},
    )
    await asyncio.sleep(10)
    assert mock_fhem.readings["testdevice"]["display_name"] == "Backdrop"

    # play servustv stream
    await fhempy_device.Set(
        testhash,
        [
            "testdevice",
            "play",
            "https://rbmn-live.akamaized.net/hls/live/2002830/geoSTVDEweb/master_6692.m3u8",
        ],
        {},
    )
    await asyncio.sleep(20)
    assert mock_fhem.readings["testdevice"]["display_name"] == "Default Media Receiver"
    assert mock_fhem.readings["testdevice"]["mediaPlayerState"] == "PLAYING"

    # volume
    await fhempy_device.Set(
        testhash,
        [
            "testdevice",
            "volume",
            "30",
        ],
        {},
    )
    await asyncio.sleep(2)
    assert mock_fhem.readings["testdevice"]["volume"] == 30

    # volume
    await fhempy_device.Set(
        testhash,
        [
            "testdevice",
            "volume",
            "20",
        ],
        {},
    )
    await asyncio.sleep(2)
    assert mock_fhem.readings["testdevice"]["volume"] == 20

    # quitApp
    await fhempy_device.Set(
        testhash,
        ["testdevice", "quitApp"],
        {},
    )
    await asyncio.sleep(10)
    assert mock_fhem.readings["testdevice"]["display_name"] == "Backdrop"



import base64
import logging
import re
import time
from telnetlib import Telnet
from typing import Union

_LOGGER = logging.getLogger(__name__)

CHECK_SOCAT = "(md5sum /data/socat | grep 92b77e1a93c4f4377b4b751a5390d979)"
DOWNLOAD_SOCAT = "(curl -o /data/socat http://pkg.musl.cc/socat/mipsel-linux-musln32/bin/socat && chmod +x /data/socat)"
RUN_SOCAT = "/data/socat tcp-l:8888,reuseaddr,fork /dev/ttyS2"

CHECK_BUSYBOX = "(md5sum /data/busybox | grep 099137899ece96f311ac5ab554ea6fec)"
DOWNLOAD_BUSYBOX = "(curl -k -o /data/busybox https://busybox.net/downloads/binaries/1.21.1/busybox-mipsel && chmod +x /data/busybox)"
LOCK_FIRMWARE = "/data/busybox chattr +i /data/firmware.bin"
UNLOCK_FIRMWARE = "/data/busybox chattr -i /data/firmware.bin"
RUN_FTP = "(/data/busybox tcpsvd -vE 0.0.0.0 21 /data/busybox ftpd -w &)"

# use awk because buffer
MIIO_LESS = "-l 0 -o FILE_STORE -n 128"
MIIO_MORE = "-l 4"
MIIO2MQTT = "(miio_client %s -d /data/miio | awk '/%s/{print $0;fflush()}' | mosquitto_pub -t log/miio -l &)"

RE_VERSION = re.compile(r'version=([0-9._]+)')


class TelnetShell(Telnet):
    def __init__(self, host: str):
        super().__init__(host, timeout=5)
        self.read_until(b"login: ")
        self.exec('admin')

    def exec(self, command: str, as_bytes=False) -> Union[str, bytes]:
        """Run command and return it result."""
        self.write(command.encode() + b"\r\n")
        raw = self.read_until(b"\r\n# ")
        return raw if as_bytes else raw.decode()

    def check_or_download_socat(self):
        """Download socat if needed."""
        return self.exec(f"{CHECK_SOCAT} || {DOWNLOAD_SOCAT}")

    def run_socat(self):
        self.exec(f"{CHECK_SOCAT} && {RUN_SOCAT} &")

    def stop_socat(self):
        self.exec(f"killall socat")

    def run_lumi_zigbee(self):
        self.exec("daemon_app.sh &")

    def stop_lumi_zigbee(self):
        self.exec("killall daemon_app.sh; killall Lumi_Z3GatewayHost_MQTT")

    def check_or_download_busybox(self):
        return self.exec(f"{CHECK_BUSYBOX} || {DOWNLOAD_BUSYBOX}")

    def check_firmware_lock(self) -> bool:
        """Check if firmware update locked. And create empty file if needed."""
        raw = self.exec("touch /data/firmware.bin")
        return "Permission denied" in raw

    def lock_firmware(self, enable: bool):
        command = LOCK_FIRMWARE if enable else UNLOCK_FIRMWARE
        self.exec(f"{CHECK_BUSYBOX} && {command}")

    def run_ftp(self):
        self.exec(f"{CHECK_BUSYBOX} && {RUN_FTP}")

    def sniff_bluetooth(self):
        self.write(b"killall silabs_ncp_bt; silabs_ncp_bt /dev/ttyS1 1\r\n")

    def run_public_mosquitto(self):
        self.exec("killall mosquitto")
        time.sleep(.5)
        self.exec("mosquitto -d")
        time.sleep(.5)
        # fix CPU 90% full time bug
        self.exec("killall zigbee_gw")

    def get_running_ps(self) -> str:
        return self.exec("ps")

    def redirect_miio2mqtt(self, pattern: str, new_version=False):
        self.exec("killall daemon_miio.sh; killall miio_client")
        time.sleep(.5)
        args = MIIO_LESS if new_version else MIIO_MORE
        self.exec(MIIO2MQTT % (args, pattern))
        self.exec("daemon_miio.sh &")

    def run_public_zb_console(self):
        self.exec("killall daemon_app.sh; killall Lumi_Z3GatewayHost_MQTT")
        # run Gateway with open console port
        self.exec("Lumi_Z3GatewayHost_MQTT -n 1 -b 115200 -v -p '/dev/ttyS2' "
                  "-d '/data/silicon_zigbee_host/' > /dev/null &")

        # connect to console to start zigbee chip
        self.write(b"nc localhost 4901\r\n")
        time.sleep(2)
        # exit console (ctrl+c)
        self.write(b"\x03")
        self.read_until(b"\r\n# ")

        self.exec("daemon_app.sh &")

    def read_file(self, filename: str, as_base64=False):
        if as_base64:
            self.write(f"cat {filename} | base64\r\n".encode())
            self.read_until(b"\r\n")  # skip command
            raw = self.read_until(b"# ")
            return base64.b64decode(raw)
        else:
            self.write(f"cat {filename}\r\n".encode())
            self.read_until(b"\r\n")  # skip command
            return self.read_until(b"# ")[:-2]

    def run_buzzer(self):
        self.exec("killall basic_gw")

    def stop_buzzer(self):
        self.exec("killall daemon_miio.sh; killall -9 basic_gw")
        # run dummy process with same str in it
        self.exec("sh -c 'sleep 999d' dummy:basic_gw &")
        self.exec("daemon_miio.sh &")

    def get_version(self):
        raw = self.read_file('/etc/rootfs_fw_info')
        m = RE_VERSION.search(raw.decode())
        return m[1]



import asyncio
import hashlib
import json
import logging
import random
import socket
import time
from asyncio.protocols import BaseProtocol
from asyncio.transports import DatagramTransport
from typing import Union, Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

_LOGGER = logging.getLogger(__name__)

HELLO = bytes.fromhex(
    "21310020ffffffffffffffffffffffffffffffffffffffffffffffffffffffff"
)


class BasemiIO:
    """A simple class that implements the miIO protocol."""
    device_id = None
    delta_ts = None

    def __init__(self, host: str, token: str):
        self.addr = (host, 54321)
        self.token = bytes.fromhex(token)

        key = hashlib.md5(self.token).digest()
        iv = hashlib.md5(key + self.token).digest()
        self.cipher = Cipher(algorithms.AES(key), modes.CBC(iv),
                             backend=default_backend())

    def _encrypt(self, plaintext: bytes):
        padder = padding.PKCS7(128).padder()
        padded_plaintext = padder.update(plaintext) + padder.finalize()

        encryptor = self.cipher.encryptor()
        return encryptor.update(padded_plaintext) + encryptor.finalize()

    def _decrypt(self, ciphertext: bytes):
        decryptor = self.cipher.decryptor()
        padded_plaintext = decryptor.update(ciphertext) + decryptor.finalize()

        unpadder = padding.PKCS7(128).unpadder()
        return unpadder.update(padded_plaintext) + unpadder.finalize()

    def _pack_raw(self, method: str, params: Union[dict, list] = None):
        # latest zero unnecessary
        payload = json.dumps({
            'id': random.randint(100000000, 999999999),
            'method': method, 'params': params or []
        }, separators=(',', ':')).encode() + b'\x00'

        data = self._encrypt(payload)

        raw = b'\x21\x31'
        raw += (32 + len(data)).to_bytes(2, 'big')  # total length
        raw += b'\x00\x00\x00\x00'  # unknow
        raw += self.device_id.to_bytes(4, 'big')
        raw += int(time.time() - self.delta_ts).to_bytes(4, 'big')

        raw += hashlib.md5(raw + self.token + data).digest()
        raw += data

        assert len(raw) < 1024, "Exceeded message size"

        return raw

    def _unpack_raw(self, raw: bytes):
        assert raw[:2] == b'\x21\x31'
        # length = int.from_bytes(raw[2:4], 'big')
        # unknown = raw[4:8]
        # device_id = int.from_bytes(raw[8:12], 'big')
        # ts = int.from_bytes(raw[12:16], 'big')
        # checksum = raw[16:32]
        return self._decrypt(raw[32:])


class SyncmiIO(BasemiIO):
    """Synchronous miIO protocol."""

    def __init__(self, host: str, token: str, timeout: float = 2):
        super().__init__(host, token)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(timeout)

    def ping(self) -> bool:
        """Returns `true` if the connection to the miio device is working. The
        token is not verified at this stage.
        """
        try:
            self.sock.sendto(HELLO, self.addr)
            raw = self.sock.recv(1024)
            if raw[:2] == b'\x21\x31':
                self.device_id = int.from_bytes(raw[8:12], 'big')
                self.delta_ts = time.time() - int.from_bytes(raw[12:16], 'big')
                return True
        except:
            pass
        return False

    def send(self, method: str, params: Union[dict, list] = None):
        """Send command to miIO device and get result from it. Params can be
        dict or list depend on command.
        """
        if not self.device_id and not self.ping():
            return None

        try:
            raw = self._pack_raw(method, params)
            self.sock.sendto(raw, self.addr)

            # can receive more than 1024 bytes
            raw = self.sock.recv(10240)
            data = self._unpack_raw(raw)

            return json.loads(data.rstrip(b'\x00'))['result']
        except Exception as e:
            _LOGGER.debug(f"Can't send: {e}")
            return None

    def send_bulk(self, method: str, params: list):
        """Sends a command with a large number of parameters. Splits into
        multiple requests when the size of one request is exceeded.
        """
        result = []
        pack = []
        total_len = 0

        try:
            for item in params:
                item_len = len(str(item))
                # approximate number, it seems to work
                if total_len + item_len > 900:
                    result += self.send(method, pack)
                    pack = []
                    total_len = 0

                pack.append(item)
                total_len += item_len

            return result + self.send(method, pack)

        except:
            return None

    def info(self):
        """Get info about miIO device."""
        return self.send('miIO.info')


class AsyncmiIO(BasemiIO, BaseProtocol):
    response = None
    sock: Optional[DatagramTransport] = None

    def datagram_received(self, data: bytes, addr):
        # hello message
        if len(data) == 32:
            if data[:2] == b'\x21\x31':
                self.device_id = int.from_bytes(data[8:12], 'big')
                ts = int.from_bytes(data[12:16], 'big')
                self.delta_ts = time.time() - ts
                result = True
            else:
                result = False
        else:
            result = self._unpack_raw(data)

        self.response.set_result(result)

    def error_received(self, exc):
        print('Error received:', exc)

    def connection_lost(self, exc):
        print("Connection closed")
        self.sock = None

    async def send_raw(self, data: bytes):
        loop = asyncio.get_running_loop()
        if not self.sock:
            self.sock, _ = await loop.create_datagram_endpoint(
                lambda: self, remote_addr=self.addr)

        self.response = loop.create_future()
        # this method does not block
        self.sock.sendto(data)
        return await self.response

    async def ping(self):
        return await self.send_raw(HELLO)

    async def send(self, method: str, params: Union[dict, list] = None):
        """Send command to miIO device and get result from it. Params can be
        dict or list depend on command.
        """
        if not self.device_id and not await self.ping():
            return None

        try:
            raw = self._pack_raw(method, params)
            data = await self.send_raw(raw)
            return json.loads(data.rstrip(b'\x00'))['result']
        except Exception as e:
            _LOGGER.warning(f"Can't send: {e}")
            return None

    async def send_bulk(self, method: str, params: list):
        """Sends a command with a large number of parameters. Splits into
        multiple requests when the size of one request is exceeded.
        """
        result = []
        pack = []
        total_len = 0

        for item in params:
            item_len = len(str(item))
            # approximate number, it seems to work
            if total_len + item_len > 900:
                result += await self.send(method, pack)
                pack = []
                total_len = 0

            pack.append(item)
            total_len += item_len

        return result + await self.send(method, pack)

    async def info(self):
        """Get info about miIO device."""
        return await self.send('miIO.info')



import asyncio
import logging
import bluetooth
import functools
from bt_proximity import BluetoothRSSI

from .. import fhem
from .. import utils

from fhempy.lib.generic import FhemModule


class bt_presence(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.btscan_task = None
        self._curr_state = ""
        self._count_diff_state = 0
        attr_config = {
            "absentInterval": {"default": 10, "format": "int"},
            "presentInterval": {"default": 60, "format": "int"},
            "absentThreshold": {"default": 0, "format": "int"},
            "presentThreshold": {"default": 0, "format": "int"},
        }
        self.set_attr_config(attr_config)

    def lookup_name(self, mac):
        return bluetooth.lookup_name(mac, timeout=5)

    async def run_bt_scan(self):
        curr_name = ""
        curr_rssi = 0
        while True:
            new_state = "absent"
            try:
                # check max 3 times for device_name
                device_name = None
                for i in range(0, 2):
                    device_name = await utils.run_blocking(
                        functools.partial(self.lookup_name, self._address)
                    )
                    if device_name:
                        break

                if device_name:
                    self._btrssi = BluetoothRSSI(self._address)
                    rssi = await utils.run_blocking(
                        functools.partial(self._btrssi.request_rssi)
                    )
                    self._btrssi.close()
                    rssi = rssi[0]
                    if curr_name != device_name:
                        curr_name = device_name
                        await fhem.readingsSingleUpdateIfChanged(
                            self.hash, "name", device_name, 1
                        )
                    if curr_rssi != rssi:
                        curr_rssi = rssi
                        await fhem.readingsSingleUpdateIfChanged(
                            self.hash, "rssi", rssi, 1
                        )
                    new_state = "present"
            except:
                self.logger.exception("BleakScanner failed")

            if self._curr_state != new_state:
                self._count_diff_state += 1
                if new_state == "present":
                    threshold = self._attr_presentThreshold
                else:
                    threshold = self._attr_absentThreshold
                if self._count_diff_state >= threshold:
                    self._curr_state = new_state
                    self._count_diff_state = 0
                    await self.update_state(self._curr_state)

            if self._curr_state == "present":
                interval = self._attr_presentInterval
            else:
                interval = self._attr_absentInterval
            await asyncio.sleep(interval)

    async def update_state(self, new_state):
        await fhem.readingsSingleUpdateIfChanged(self.hash, "presence", new_state, 1)
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", new_state, 1)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 4:
            return "Usage: define p_mysmartphone PythonModule bt_presence <MAC>"

        self._address = args[3]
        self.hash["MAC"] = args[3]

        if self.btscan_task:
            self.btscan_task.cancel()
        self.btscan_task = self.create_async_task(self.run_bt_scan())
